from typing import Sequence, Hashable, Union

from hypothesis import given, infer, assume

from collectionish import UniqueTuple
from collectionish.utils import is_hashable

from tests.utils import param, mark_params


@given(args=infer)
def test_uniquetuple_is_unique(args: Sequence[Hashable]):
    assume(all(map(is_hashable, args)))
    assert len(UniqueTuple(*args)) == len(set(args))


@mark_params
@param(tag='ordering', inp=(3, 3, 2, 3, 2, 1, 2, 3), expected=UniqueTuple(3, 2, 1))
@param(tag='empty', inp=[], expected=UniqueTuple())
def test_explicit_examples(inp, expected):
    assert UniqueTuple(*inp) == expected


@mark_params
@param(inp=int)
@param(inp=str)
@param(inp=Union[int, str])
def test_typehints(inp):
    hint = UniqueTuple[inp]
    assert hint.__args__ == (inp,)
    assert hint.__origin__ == UniqueTuple



import pytest


def mark_params(f):
    return pytest.mark.parametrize(f._argnames, f._params)(f)


def param(tag=None, **kwargs):
    def wrapper(f):
        seen = hasattr(f, '_argnames') and hasattr(f, '_params')
        if not seen:
            f._argnames = tuple(kwargs)
            f._params = []
        p = pytest.param(*[kwargs[k] for k in f._argnames], id=tag)
        f._params.append(p)
        return f

    return wrapper



import typing as t
import ast
from types import MappingProxyType

import hypothesis.strategies as st
from hypothesis import given

from collectionish.utils import is_valid_identifier
from collectionish.utils import is_mapping, is_arraylike

from tests.utils import param, mark_params
from tests.strategies import everything_except, one_of_types

MAPPING_TYPES = (t.Mapping, t.MutableMapping, dict)


def validate_identifier_as_ast(s: str):
    parsed = ast.parse(f'{s}=1', mode='single')
    assert len(parsed.body) == 1
    body = parsed.body[0]
    assert isinstance(body, ast.Assign)
    assert body.lineno == 1
    assert body.col_offset == 0

    assert body.value.n == 1

    assert len(body.targets) == 1
    name = body.targets[0]
    assert isinstance(name, ast.Name)
    assert name.id == s


@given(st.text())
def test_is_valid_identifier(s):
    if is_valid_identifier(s):
        validate_identifier_as_ast(s)


@mark_params
@param(obj=one_of_types(*MAPPING_TYPES) | st.from_type(dict).map(MappingProxyType), expected=True)
@param(obj=everything_except(*MAPPING_TYPES, MappingProxyType), expected=False)
@given(data=st.data())
def test_is_mapping(data, obj, expected):
    inp = data.draw(obj)
    assert is_mapping(inp) == expected


@mark_params
@param(obj=one_of_types(list, tuple, frozenset, set), expected=True)
@param(obj=everything_except(list, tuple, frozenset, set), expected=False)
@given(data=st.data())
def test_is_arraylike(data, obj, expected):
    inp = data.draw(obj)
    assert is_arraylike(inp) == expected



from typing import Any, Collection, Mapping

import keyword


def is_valid_identifier(s: str) -> bool:
    """determines if a str is valid python identifier.
    """
    return s.isidentifier() and not keyword.iskeyword(s)


def is_mapping(obj: Any) -> bool:
    """Just a shorthand for instancecheck on :py:class:`typing.Mapping`.
    """
    return isinstance(obj, Mapping)


def is_arraylike(obj: Any) -> bool:
    """Determine if the provided object is an non-mapping generic container type such as an array, tuple or set.

    Example:
        >>> from collectionish.utils import is_arraylike
        >>>
        >>> is_arraylike([1, 2, 3])
        True
        >>> is_arraylike(set([1, 2, 3]))
        True
        >>> is_arraylike(tuple([1, 2, 3]))
        True
        >>> is_arraylike({'a': 1, 'b': 2})
        False
        >>> is_arraylike('a string')
        False
    """
    exclude = (Mapping, str, bytes, bytearray, range, memoryview)
    return isinstance(obj, Collection) and not isinstance(obj, exclude)


def is_hashable(obj: Any) -> bool:
    """Determine if the given object is hashable.

    Example:
        >>> from collectionish.utils import is_hashable
        >>>
        >>> is_hashable([1, 2, 3])
        False
        >>> is_hashable('boop')
        True
    """
    try:
        hash(obj)
        return True
    except TypeError:
        return False



# -*- coding: utf-8 -*-

"""collectionish is a repository of possibly useful, fairly generic container types.
"""

__author__ = """Lea Provenzano"""
__email__ = 'leaprovenzano@gmail.com'
__version__ = '0.1.1'

# this is exists to make sphinx happy
__all__ = ['UniqueTuple', 'AttyDict']


from ._uniquetuple import UniqueTuple
from ._attydict import AttyDict



from typing import Dict, TypeVar

from collectionish.utils import is_valid_identifier

T = TypeVar('T')


class AttyDict(Dict[str, T]):

    """A lightweight dictionary with dot access.

    ``attydicts`` are dictionaries witch allow access to values values through dot notation.
    The limitation there is that their keys must be strings ( and valid python identifiers )
    but sometimes this is what you want when working with json objects unpredictable json
    or the like. That said, if you have sturctured data you'd  of course be much better off \
    using a structured datatype.

    Example:

        initilize a basic attydict:

        >>> from collectionish import AttyDict
        >>>
        >>> the_sea = AttyDict(crabs=10, fish=2)
        >>> the_sea
        {'crabs': 10, 'fish': 2}

        access stuff with dots:

        >>> the_sea.crabs += 1
        >>> the_sea.crabs
        11

        nested stuff will also have attr access:

        >>> the_sea.update(submarines= {'sandwich': 0, 'actual': 1})
        >>> the_sea
        {'crabs': 11, 'fish': 2, 'submarines': {'sandwich': 0, 'actual': 1}}
        >>> the_sea.submarines.actual
        1

        we handle name clashes in a similar way to pandas - Reserved names can be set using standard dict access:

        >>> the_sea['pop'] = 'corn'
        >>> the_sea
        {'crabs': 11, 'fish': 2, 'submarines': {'sandwich': 0, 'actual': 1}, 'pop': 'corn'}
        >>> the_sea.pop('pop')
        'corn'
        >>> the_sea
        {'crabs': 11, 'fish': 2, 'submarines': {'sandwich': 0, 'actual': 1}}

        and as youd expect lists containing dictionaries just work...

        >>> the_sea.whales = [{'killer': 1}, {'humpback': 2}, 'mr. whale']
        >>> the_sea.whales[1].humpback
        2
    """

    @classmethod
    def _attrify(cls, value):
        if hasattr(value, '__iter__'):
            if isinstance(value, (list, tuple)):
                return type(value)((cls._attrify(v) for v in value))
            if isinstance(value, dict) and not isinstance(value, cls):
                return cls(**value)
        return value

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            self.__setitem__(k, v)

    def _validate_key(self, s: str):
        if not isinstance(s, str):
            raise TypeError(f'{self.__class__.__name__} keys must be strings.')
        if not is_valid_identifier(s):
            raise SyntaxError(f'{self.__class__.__name__} keys must be valid python identifiers.')

    def __getattr__(self, key: str) -> T:
        return self[key]

    def __setattr__(self, key: str, value: T):
        self[key] = value

    def __setitem__(self, key: str, value: T):
        if key not in self:
            self._validate_key(key)
        dict.__setitem__(self, key, self._attrify(value))

    def update(self, **kwargs):
        for k, v in kwargs.items():
            self[k] = v



from typing import NamedTuple, Any, Iterator, Tuple, Dict


import string
from types import SimpleNamespace

import pytest

import hypothesis.strategies as st
from hypothesis import given

from collectionish.utils import is_valid_identifier
from collectionish.utils import is_mapping


from collectionish.ops import rgetattr

py_names = st.text(string.ascii_lowercase + '_').filter(is_valid_identifier)


class Pair(NamedTuple):

    path: Tuple[str]
    value: Any


class Thing(SimpleNamespace):

    @classmethod
    def from_dict(cls, dct: Dict[str, Any]) -> 'Thing':
        kwargs = {}
        for k, v in dct.items():
            if is_mapping(v):
                kwargs[k] = cls.from_dict(v)
            else:
                kwargs[k] = v
        return cls(**kwargs)

    def paths(self, prefix=tuple([])) -> Iterator[Pair]:
        yield Pair(prefix, self)
        for k, v in self.__dict__.items():
            if isinstance(v, self.__class__):
                for path, v in v.paths(prefix + (k,)):
                    yield Pair(path, v)
            else:
                yield Pair(prefix + (k,), v)


things = st.builds(
    Thing.from_dict, st.dictionaries(keys=py_names, values=st.text(string.ascii_lowercase))
)


@given(things)
def test_rgetattr(thing):
    for path, value in thing.paths():
        assert rgetattr(thing, *path) == value


@given(things, st.data())
def test_rgetattr_raises_on_invalid(thing, data):
    path, _ = data.draw(st.sampled_from(list(thing.paths())))
    with pytest.raises(AttributeError):
        rgetattr(thing, *path, 'bullshit')



from typing import Union, Mapping
from copy import deepcopy
import string

from hypothesis import given, assume
from hypothesis import strategies as st
from hypothesis.stateful import (
    RuleBasedStateMachine,
    Bundle,
    rule,
    precondition,
    consumes,
    invariant,
)


from collectionish import AttyDict
from collectionish import ops

from collectionish.utils import is_valid_identifier

from tests.utils import param, mark_params


valid_keys = st.text(string.ascii_lowercase + '_').filter(is_valid_identifier)
flat_values = st.text(string.ascii_lowercase).filter(is_valid_identifier)
nested_values = st.recursive(
    flat_values,
    lambda children: st.one_of(
        st.lists(children, min_size=1, max_size=5),
        st.dictionaries(valid_keys, children, min_size=1, max_size=5),
        st.dictionaries(valid_keys, children, min_size=1, max_size=5).map(lambda x: AttyDict(**x)),
    ),
    max_leaves=5,
)


def valid_values(*values) -> bool:
    def valid_value(value):
        if isinstance(v, Mapping):
            if not isinstance(v, AttyDict):
                return False
            return valid_values(value.values())
        elif isinstance(v, (list, tuple, set)):
            return valid_values(*v)
        return True

    for v in values:
        if not valid_value(v):
            return False
    return True


@mark_params
@param(inp=int)
@param(inp=str)
@param(inp=Union[int, str])
def test_typehints(inp):
    hint = AttyDict[inp]
    assert hint.__args__ == (inp,)
    assert hint.__origin__ == AttyDict


@mark_params
@param(tag='flat', inp={'b': 1})
@param(tag='nested', inp={'a': {'aa': 2}})
@param(tag='deep_nested', inp={'a': {'aa': {'aaa': 1}}})
@param(tag='mixed', inp={'a': {'aa': 2}, 'b': 2})
def test_update(inp):
    """Test an AttyDict update works just the same as a dict update
    """
    atty = AttyDict(a={'aa': 1, 'ab': 2})
    regular = dict(a={'aa': 1, 'ab': 2})

    atty.update(**inp)
    assert valid_values(atty)

    regular.update(**inp)
    assert dict(atty) == regular


@given(st.dictionaries(valid_keys, nested_values))
def test_init(inp):
    atty = AttyDict(**inp)
    assert valid_values(atty)
    assert dict(atty) == inp


class AttyDictMachine(RuleBasedStateMachine):

    def __init__(self):
        super().__init__()
        self.atty = AttyDict()
        self.shadow = dict()
        self.change = 0
        print()

    keys = Bundle('keys')
    values = Bundle('values')

    @rule(target=keys, k=valid_keys)
    def add_new_key(self, k):
        return k

    @precondition(
        lambda self: len([k for k, v in self.atty.items() if isinstance(v, AttyDict)]) > 0
    )
    @rule(target=keys, existing=st.data(), k=valid_keys)
    def add_nested_key(self, existing, k):
        ex = existing.draw(
            st.sampled_from([k for k, v in self.atty.items() if isinstance(v, AttyDict)])
        )
        return (ex, k)

    @rule(target=values, v=flat_values)
    def add_flat_value(self, v):
        return v

    @rule(target=values, v=st.dictionaries(keys=valid_keys, values=flat_values))
    def add_dict_values(self, v):
        return v

    @rule(k=consumes(keys), v=consumes(values))
    def set_as_attr(self, k, v):
        if isinstance(k, tuple):
            assume(k[0] in self.atty)
            ops.rsetattr(self.atty, k, deepcopy(v))
            ops.rsetitem(self.shadow, k, deepcopy(v))
        else:
            setattr(self.atty, k, deepcopy(v))
            self.shadow[k] = deepcopy(v)
        self.change = 1
        return (k, v)

    @rule(k=consumes(keys), v=consumes(values))
    def set_as_key(self, k, v):
        if isinstance(k, tuple):
            assume(k[0] in self.atty)
            ops.rsetitem(self.atty, k, v)
            ops.rsetitem(self.shadow, k, deepcopy(v))
        else:
            self.atty[k] = v
            self.shadow[k] = deepcopy(v)
        self.change = 1
        return (k, v)

    @precondition(lambda self: self.change == 1)
    @invariant()
    def check(self):
        self.change = 0
        assert valid_values(*self.atty.values())
        assert dict(self.atty) == self.shadow


TestAttyDict = AttyDictMachine.TestCase



from typing import Type, Iterable
from hypothesis import strategies as st


def everything_except(*excluded_types: Iterable[Type]) -> st.SearchStrategy:
    """strategy which returns everything except the excluded types.
    """
    return (
        st.from_type(type).flatmap(st.from_type).filter(lambda x: not isinstance(x, excluded_types))
    )


def one_of_types(*types: Iterable[Type]) -> st.SearchStrategy:
    """return union of type strategies over multiple types.
    """
    return st.one_of(st.from_type(t) for t in types)



import random

from collectionish import UniqueTuple

n = 1000
unique_unordered_ints = list(range(n))
random.shuffle(unique_unordered_ints)
repeated_unordered_ints = [random.choice(unique_unordered_ints[:100]) for i in range(n)]


def test_benchmark_uniquetuple1(benchmark):

    benchmark.group = 'UniqueTuple.unique_unordered'

    def f():
        lambda: UniqueTuple(*unique_unordered_ints)

    return benchmark(f)


def test_benchmark_uniquetuple2(benchmark):

    benchmark.group = 'UniqueTuple.repeated_unordered'

    def f():
        return UniqueTuple(*repeated_unordered_ints)

    return benchmark(f)



from typing import Any, Iterable, Sequence, Hashable, MutableMapping, Mapping, Union
from functools import reduce


def rgetattr(obj: Any, *keys: Iterable[str]):
    """A recursive version of :py:func:`getattr`.

    Example:

        Here's a simple example using :py:class:`types.SimpleNamespace` as a standin for \
        anything with attibute access:

        >>> from types import SimpleNamespace
        >>> from collectionish.ops import rgetattr
        >>>
        >>> thing = SimpleNamespace(a=1, b=SimpleNamespace(ba=1, bb=SimpleNamespace(bba=1, bbb=2)))

        with only one key it works like normal :py:func:`getattr`:

        >>> rgetattr(thing, 'a')
        1
        >>> rgetattr(thing, 'b')
        namespace(ba=1, bb=namespace(bba=1, bbb=2))

        with multiple ``rgetattr`` gathers the attibute from a nested object:

        >>> rgetattr(thing, 'b', 'ba')
        1
        >>> rgetattr(thing, 'b', 'bb', 'bbb')
        2

    """
    return reduce(getattr, keys, obj)  # type: ignore


def rsetattr(obj: Any, keys: Iterable[str], value: Any):
    """A recursive version of :py:func:`setattr`.

    Example:

        Again we'll use  :py:class:`types.SimpleNamespace` as a standin for \
        anything with attibute access:

        >>> from types import SimpleNamespace
        >>> from collectionish.ops import rgetattr
        >>>
        >>> thing = SimpleNamespace()

        with only one key it works like normal :py:func:`setattr`:

        >>> rsetattr(thing, ['a'], 1)
        >>> thing
        namespace(a=1)
        >>> rsetattr(thing, ['b'], SimpleNamespace(ba= 1, bb= 2))
        >>> thing
        namespace(a=1, b=namespace(ba=1, bb=2))

        with multiple ``rsetattr`` can update a nested value:

        >>> rsetattr(thing, ('b', 'ba'), 2)
        >>> thing
        namespace(a=1, b=namespace(ba=2, bb=2))

    """
    *keys, last = keys
    parent = rgetattr(obj, *keys)
    setattr(parent, last, value)


def rgetitem(obj: Union[Sequence, Mapping], *keys: Iterable[Hashable]):
    """A recursive version of :py:func:`__getitem__`.

    Example:

        >>> from collectionish.ops import rgetitem
        >>>
        >>> thing = {'a': 1, 'b': {'ba': 1, 'bb': 2}}

        with only one key it works like normal :py:func:`__getitem__`:

        >>> rgetitem(thing, 'a')
        1

        with multiple ``rgetattr`` gets items recursively:

        >>> rgetitem(thing, 'b', 'bb')
        2

        works fine with lists and stuff as well:

        >>> nested_list = [1, [1, 2, [1, 2, 3]], 2]
        >>> rgetitem(nested_list, -1)
        2
        >>> rgetitem(nested_list, 1, 2, 2)
        3
    """
    return reduce(lambda x, k: x[k], keys, obj)  # type: ignore


def rsetitem(obj: Union[Sequence, MutableMapping], keys: Iterable[Hashable], value: Any) -> None:
    """A recursive version of :py:func:`__setitem__`.

    Example:

        >>> from collectionish.ops import rgetitem
        >>>
        >>> thing = {'a': 1, 'b': {'ba': 1, 'bb': 2}}

        with only one key it works like normal :py:func:`__setitem__`:

        >>> rsetitem(thing, ('a',), 2)
        >>> thing
        {'a': 2, 'b': {'ba': 1, 'bb': 2}}

        with multiple ``rgetattr`` gets items recursively:

        >>> rsetitem(thing, ['b', 'bb'], 4)
        >>> thing
        {'a': 2, 'b': {'ba': 1, 'bb': 4}}

        lists and stuff work as expected:

        >>> nested_list = [1, [1, 2, [1, 2, 3]], 2]
        >>> rsetitem(nested_list, [-1], 3)
        >>> nested_list
        [1, [1, 2, [1, 2, 3]], 3]
        >>> rsetitem(nested_list, [1, 2, 2], 5)
        >>> nested_list
        [1, [1, 2, [1, 2, 5]], 3]

    """
    *keys, last = keys
    parent = rgetitem(obj, *keys)  # type: ignore
    parent[last] = value



from typing import Tuple, TypeVar, Hashable, Sequence


T = TypeVar('T', bound=Hashable, covariant=True)


class UniqueTuple(Tuple[T, ...]):

    """An immutable sequence of unique and hashable items ordered by first appearance.

    - like ``tuple`` and ``frozenset``,  ``UniqueTuple`` is immutable.
    - like ``frozenset``, ``UniqueTuple`` accepts only hashable values.
    - `unlike` ``frozenset``, ``UniqueTuple`` cares about insertion order. \
    It orders by will be ordered by first appearance.
    - finally ``UniqueTuple`` may be initilized with an unpacked iterable.

    Example:

        >>> from collectionish import UniqueTuple
        >>>
        >>> UniqueTuple(3, 2, 3, 1)
        UniqueTuple(3, 2, 1)

    """

    def __new__(cls, *args: Sequence[T]):
        # since dict remembers insertion order we can just use that with no values
        return super().__new__(cls, dict({arg: None for arg in args}))  # type: ignore

    def count(self, value) -> int:
        return int(value in self)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}{super().__repr__()}'
